#ifndef AG_ALEATORY_BINARY_TREE_H
#define AG_ALEATORY_BINARY_TREE_H

/*
 * The AGCLIB1 Class Library, Version 1.0
 * Copyright (c) Parsifal Software, 2001. All Rights Reserved.
 * http://www.parsifalsoft.com
*/

#include <stdlib.h>
#include <assert.h>

int agABTHash(const void *v, int n, int startValue = 0);
inline int agABTHash(int x, int startValue = 0) {return agABTHash(&x, sizeof(int), startValue);}
inline int agABTHash(double x, int startValue = 0) {return agABTHash(&x, sizeof(double), startValue);}

template <class Container>
int agABTHashContainer(const Container &c, int hash) {
  int k = 5;
  int n = c.size();
  if (k > n) k = n;
  int delta = n/k;
  int i = hash;
  while (k--) {
    hash = agABTHash(c[i%n], hash);
    i += delta;
  }
  return hash;
}

template <class Container>
int agABTHashContainer(const Container &c) {
  return agABTHashContainer(c, 0);
}

template <class Object>
struct AgAleatoryBinaryNode {
  Object object;
  AgAleatoryBinaryNode<Object> *left, *right;                              // subnodes
  int size;
  int splitter;                                   // randomizes subnodes
  AgAleatoryBinaryNode(const Object &e);
  AgAleatoryBinaryNode(const AgAleatoryBinaryNode<Object> &n);
  ~AgAleatoryBinaryNode();
  operator Object &() {return object;}
  operator const Object &() const {return object;}
  AgAleatoryBinaryNode<Object> &find(int i);
  void setSize();
  int nextHashcode(int hashcode) const {
    return (char)(splitter + 2*hashcode)*257 >> 4;
  }
  int hash() const;

#ifdef TESTING_AG_SET
  int maxDepth() const {
    int d = left ? left->maxDepth() : 0;
    int dr = right ? right->maxDepth() : 0;
    if (dr > d) d = dr;
    d++;
    return d;
  }
  int sumDepth(int depth) const {
    depth++;
    int d = left ? left->sumDepth(depth) : depth;
    if (right) d += right->sumDepth(depth);
    return d;
  }
#endif
};

template <class Object>
AgAleatoryBinaryNode<Object>::AgAleatoryBinaryNode(const Object &e)
  : object(e), left(NULL), right(NULL), size(1),
    splitter((rand() >> 4))
{
  // Nothing to do
}

template <class Object>
AgAleatoryBinaryNode<Object>::AgAleatoryBinaryNode(const AgAleatoryBinaryNode<Object> &n)
  : object(n.object),
    left(n.left ? new AgAleatoryBinaryNode<Object>(*n.left) : NULL),
    right(n.right ? new AgAleatoryBinaryNode<Object>(*n.right) : NULL),
    size(n.size),
    splitter(n.splitter)
{
  // Nothing to do
}

template <class Object>
AgAleatoryBinaryNode<Object>::~AgAleatoryBinaryNode() {
  delete left;
  delete right;
}

template <class Object>
AgAleatoryBinaryNode<Object> &AgAleatoryBinaryNode<Object>::find(int i) {
  int leftSize = left ? left->size : 0;
  return i < leftSize ? left->find(i)
          : i == leftSize ? *this
          : right->find(i - leftSize - 1);
}

template <class Object>
void AgAleatoryBinaryNode<Object>::setSize() {
  size = (left ? left->size : 0) + (right ? right->size : 0) + 1;
}

template <class Object>
int AgAleatoryBinaryNode<Object>::hash() const {
  int hashcode = agABTHash(object);
  if (left) hashcode ^= left->hash();
  if (right) hashcode ^= right->hash();
  return hashcode;
}

template <class Object>
class AgAleatoryBinaryTree {
protected:
  typedef AgAleatoryBinaryNode<Object> Node;
  Node *root;

  Node *&find(Node *const &, const Object &, int hashcode) const;
  Node  *findOrCreate(Node *&, const Object &, int hashcode);
  int    findAndRemove(Node *&, const Object &, int hashcode);
  Node *removeDeepestLeaf(Node *&node);
  virtual int equivalent(const Object &x, const Object &y) const = 0;
  AgAleatoryBinaryTree() : root(NULL) {}
  AgAleatoryBinaryTree(const AgAleatoryBinaryTree<Object> &s);
  AgAleatoryBinaryTree<Object> &operator =(const AgAleatoryBinaryTree<Object> &s);
  virtual ~AgAleatoryBinaryTree() {delete root;}
public:
  AgAleatoryBinaryTree<Object> &reset();
  int size() const {return root ? root->size : 0;}
  int hash() const {return root ? agABTHash(root->size, root->hash()) : 0;}
#ifdef TESTING_AG_SET
  int maxDepth() const {return root ? root->maxDepth() : 0;}
  double avgDepth() const {return root ? (double)root->sumDepth(0)/size() : 0;}
#endif
};

// Copy constructor
template <class Object>
AgAleatoryBinaryTree<Object>::AgAleatoryBinaryTree(const AgAleatoryBinaryTree<Object> &s)
: root(s.root ? new Node(*s.root) : NULL)
{
  // Nothing to do
}

// Assignment operator
template <class Object>
AgAleatoryBinaryTree<Object> &AgAleatoryBinaryTree<Object>::operator =(const AgAleatoryBinaryTree<Object> &s) {
  if (&s == this) return *this;
  delete root;
  root = s.root ? new Node(*s.root) : NULL;
  return *this;
}

// reset function
template <class Object>
AgAleatoryBinaryTree<Object> &AgAleatoryBinaryTree<Object>::reset() {
  delete root;
  root = NULL;
  return *this;
}

// find a specified object
template <class Object>
AgAleatoryBinaryNode<Object> *&AgAleatoryBinaryTree<Object>::find(
    Node *const &node, const Object &object, int hashcode) const
{
  if (node == NULL || equivalent(object, node->object)) return (Node *&) node;
  int newHashcode = node->nextHashcode(hashcode);
  return (signed char)(node->splitter - hashcode) < 0
         ? find(node->left, object, newHashcode)
         : find(node->right, object, newHashcode);
}

// find an object, add it if it's not already there
template <class Object>
AgAleatoryBinaryNode<Object> *AgAleatoryBinaryTree<Object>::findOrCreate(
    Node *&node, const Object &object, int hashcode)
{
  if (node == NULL) {
    node = new Node(object);
    return node;
  }
  if (equivalent(object, node->object)) return node;
  int newHashcode = node->nextHashcode(hashcode);
  Node *returnValue = (signed char)(node->splitter - hashcode) < 0
                    ? findOrCreate(node->left, object, newHashcode)
                    : findOrCreate(node->right, object, newHashcode);
  node->setSize();
  return returnValue;
}

// find an object and remove it if it's there
template <class Object>
int AgAleatoryBinaryTree<Object>::findAndRemove(
    Node *&node, const Object &object, int hashcode)
{
  if (node == NULL) return 0;
  if (equivalent(object, node->object)) {
    Node *leaf = removeDeepestLeaf(node);
    if (node == NULL) delete leaf;
    else {
      leaf->left = node->left; leaf->right = node->right;
      leaf->size = node->size; leaf->splitter = node->splitter;
      node->left = node->right = NULL;
      delete node;
      node = leaf;
    }
    return 1;
  }
  int newHashcode = node->nextHashcode(hashcode);
  int returnValue = (signed char)(node->splitter - hashcode) < 0
                      ? findAndRemove(node->left, object, newHashcode)
                      : findAndRemove(node->right, object, newHashcode);
  node->setSize();
  return returnValue;
}

// Support function for findAndRemove
template <class Object>
AgAleatoryBinaryNode<Object> *AgAleatoryBinaryTree<Object>::removeDeepestLeaf(Node *&node) {
  Node *returnValue = node;
  if (node->left) {
    returnValue = node->right && node->right->size > node->left->size
                  ? removeDeepestLeaf(node->right) : removeDeepestLeaf(node->left);
    if (node) node->setSize();
  }
  else if (node->right) {
    returnValue = removeDeepestLeaf(node->right);
    if (node) node->setSize();
  }
  else node = NULL;
  return returnValue;
}

#endif
